// Copyright (c) 2011 Sirikata Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can
// be found in the LICENSE file.

#ifndef _SIRIKATA_OH_OBJECT_QUERY_PROCESSOR_HPP_
#define _SIRIKATA_OH_OBJECT_QUERY_PROCESSOR_HPP_

#include <sirikata/oh/Platform.hpp>
#include <sirikata/oh/HostedObject.hpp>
#include <sirikata/core/service/Service.hpp>
#include <sirikata/core/util/Factory.hpp>

namespace Sirikata {

namespace Protocol {
namespace Prox {
class ProximityResults;
}
namespace Loc {
class BulkLocationUpdate;
}
}

namespace OH {

/** ObjectQueryProcessor is the interface for a geometric (proximity) query
 *  service for objects in an object host. The query format is opaque: different
 *  implementations may support different query formats. Results are provided
 *  directly to the HostedObject in the form of
 *  Protocol::Prox::ProximityResults, as sets of atomic additions and
 *  removals. An implementation may generate results in different ways -- simple
 *  versions might just forward requests to the server and manage receiving
 *  results whereas more complicated versions might process queries
 *  locally.
 *
 *  ObjectQueryProcessor is also responsible for managing location
 *  updates. This is necessary because proximity and location are
 *  tightly coupled -- their updates depend on synchronized sequence
 *  numbers and updates for objects may not even be directly from the
 *  server (i.e. they may be generated by local query processing).
 */
class SIRIKATA_OH_EXPORT ObjectQueryProcessor : public Service {
  public:
    typedef HostedObject::SSTStreamPtr SSTStreamPtr;
    typedef HostedObject::SSTConnectionPtr SSTConnectionPtr;

    virtual ~ObjectQueryProcessor();

    // Connection/disconnection to track unrequested

    /** Invoked by the ObjectHost when an object's presence is
     *  connected to the space. While it has received an
     *  acknowledgement, the connection may not fully be established
     *  (e.g. streams may not have finished connecting).
     */
    virtual void presenceConnected(HostedObjectPtr ho, const SpaceObjectReference& sporef);
    /** Invoked by the ObjectHost when an object's presence is
     *  connected to the space and the underlying stream has been
     *  fully established.
     */
    virtual void presenceConnectedStream(HostedObjectPtr ho, const SpaceObjectReference& sporef, SSTStreamPtr strm);
    /** Invoked by the ObjectHost when an object's presence has been
     *  disconnected from the space.
     */
    virtual void presenceDisconnected(HostedObjectPtr ho, const SpaceObjectReference& sporef);


    /** Request a query during connection. This is used by the object host to
     *  inform the ObjectQueryProcessor that the object wants to connect with
     *  the given query. This gives a chance to filter the query before it is
     *  passed to the space. However, the query processor should *not* start
     *  processing it until the object has been successfully connected to the
     *  space. The default implementation is a noop.
     *
     *  \param ho the HostedObject requesting connection
     *  \param sporef the ID of the presence requesting connection
     *  \param query the query requested
     *  \returns a new query to actually pass to the space server with the
     *           connection request. It may be empty if no object-specific query
     *           should be registered
     */
    virtual String connectRequest(HostedObjectPtr ho, const SpaceObjectReference& sporef, const String& query);

    /** Request an update to the query parameters for the specified presence.
     *  \param ho the HostedObject requesting the update
     *  \param sporef the ID of the presence to update a query for
     *  \param new_query the encoded, updated query
     */
    virtual void updateQuery(HostedObjectPtr ho, const SpaceObjectReference& sporef, const String& new_query) = 0;

  protected:
    /** Helper method for implementations which delivers proximity
     *  results to the HostedObject.
     *  \param ho the HostedObject requesting the update
     *  \param sporef the ID of the presence that registered the query
     *  \param update a single proximity update to apply
     */
    void deliverProximityUpdate(HostedObjectPtr ho, const SpaceObjectReference& sporef, const Sirikata::Protocol::Prox::ProximityUpdate& update);

    /** Helper method for implementations which delivers location
     *  updates to the HostedObject.
     *  \param ho the HostedObject requesting the update
     *  \param sporef the ID of the presence that registered the query
     *  \param lu the location update
     */
    void deliverLocationUpdate(HostedObjectPtr ho, const SpaceObjectReference& sporef, const LocUpdate& lu);
};


class SIRIKATA_OH_EXPORT ObjectQueryProcessorFactory
    : public AutoSingleton<ObjectQueryProcessorFactory>,
      public Factory2<ObjectQueryProcessor*, ObjectHostContext*, const String&>
{
public:
    static ObjectQueryProcessorFactory& getSingleton();
    static void destroy();
};

} // namespace OH
} // namespace Sirikata

#endif //_SIRIKATA_OH_OBJECT_QUERY_PROCESSOR_HPP_
